// -*- C++ -*-
//
// GenericWidthGenerator.cc is a part of Herwig++ - A multi-purpose Monte Carlo event generator
// Copyright (C) 2002-2007 The Herwig Collaboration
//
// Herwig++ is licenced under version 2 of the GPL, see COPYING for details.
// Please respect the MCnet academic guidelines, see GUIDELINES for details.
//
//
// This is the implementation of the non-inlined, non-templated member
// functions of the GenericWidthGenerator class.
//

#include "GenericWidthGenerator.h"
#include "ThePEG/Interface/ClassDocumentation.h"
#include "ThePEG/Interface/Switch.h"
#include "ThePEG/Interface/Parameter.h"
#include "ThePEG/Interface/ParVector.h"
#include "TwoBodyAllOnCalculator.h"
#include "OneOffShellCalculator.h"
#include "TwoOffShellCalculator.h"
#include "ThePEG/Persistency/PersistentOStream.h"
#include "ThePEG/Persistency/PersistentIStream.h"
#include "ThePEG/Repository/Repository.h"
#include "ThePEG/Utilities/Throw.h"
#include "ThePEG/Utilities/StringUtils.h"

using namespace Herwig;

void GenericWidthGenerator::persistentOutput(PersistentOStream & os) const {
  os << _theParticle << ounit(_mass,GeV) << _prefactor << _MEtype << _MEcode
     << ounit(_MEmass1,GeV) << ounit(_MEmass2,GeV) << _MEcoupling << _modeon
     << ounit(_intermasses,GeV) << ounit(_interwidths,GeV) 
     << _noofentries << _initialize << _BRnorm
     << _npoints << _decaymodes << _decaytags << ounit(_minmass,GeV) 
     << _BRminimum << _intorder;
}

void GenericWidthGenerator::persistentInput(PersistentIStream & is, int) {
  is >> _theParticle >> iunit(_mass,GeV) >> _prefactor >> _MEtype >> _MEcode 
     >> iunit(_MEmass1,GeV) >> iunit(_MEmass2,GeV) >> _MEcoupling >>_modeon
     >> iunit(_intermasses,GeV) >> iunit(_interwidths,GeV)
     >> _noofentries >> _initialize >> _BRnorm
     >> _npoints >> _decaymodes >> _decaytags >> iunit(_minmass,GeV)
     >> _BRminimum >> _intorder;
}

ClassDescription<GenericWidthGenerator> GenericWidthGenerator::initGenericWidthGenerator;
// Definition of the static class description member.

void GenericWidthGenerator::setParticle(string p) {
  _theParticle = Repository::findParticle(StringUtils::basename(p));
  if ( ! _theParticle ) 
    Throw<InterfaceException>() 
      << "Could not set Particle interface "
      << "for the object \"" << name()
      << "\". Particle \"" << StringUtils::basename(p) << "\" not found."
      << Exception::runerror;
}

string GenericWidthGenerator::getParticle() const {
  return _theParticle ? _theParticle->fullName() : "";
}

void GenericWidthGenerator::Init() {

  static ClassDocumentation<GenericWidthGenerator> documentation
    ("The GenericWidthGenerator class is the base class for running widths");

  static Parameter<GenericWidthGenerator,string> interfaceParticle
    ("Particle",
     "The particle for which this is the width generator",
     0, "", true, false,
     &GenericWidthGenerator::setParticle, 
     &GenericWidthGenerator::getParticle);

  static Switch<GenericWidthGenerator,bool> interfaceInitialize
    ("Initialize",
     "Initialize the width using the particle data object",
     &GenericWidthGenerator::_initialize, false, false, false);
  static SwitchOption interfaceInitializeInitialization
    (interfaceInitialize,
     "Yes",
     "Do the initialization",
     true);
  static SwitchOption interfaceInitializeNoInitialization
    (interfaceInitialize,
     "No",
     "Don't do the initalization",
     false);

  static ParVector<GenericWidthGenerator,int> interfacemetype
    ("MEtype",
     "The type of matrix element either 2-body from this class or higher from"
     " class inheriting from this",
     &GenericWidthGenerator::_MEtype,
     0, 0, 0, 0, 3, false, false, true);

  static ParVector<GenericWidthGenerator,int> interfacemecode
    ("MEcode",
     "The code of matrix element either 2-body from this class or higher from"
     " class inheriting from this",
     &GenericWidthGenerator::_MEcode,
     0, 0, 0, -1, 200, false, false, true);

  static ParVector<GenericWidthGenerator,Energy> interfaceMinimumMasses
    ("MinimumMasses",
     "The minimum mass of the decay products",
     &GenericWidthGenerator::_minmass,
     GeV, 0, 0*GeV, 0*GeV,  1.E12*GeV, false, false, true);

  static ParVector<GenericWidthGenerator,double> interfaceMEcoupling
    ("MEcoupling",
     "The coupling for a given ME",
     &GenericWidthGenerator::_MEcoupling,
     0, 0, 0, 0, 1.E12, false, false, true);

  static ParVector<GenericWidthGenerator,bool> interfaceModeOn
    ("ModeOn",
     "Is this mode included in the total width calculation",
     &GenericWidthGenerator::_modeon,
     0, 0, 0, 0, 1, false, false, true);

  static ParVector<GenericWidthGenerator,Energy> interfaceMEmass1
    ("MEmass1",
     "The mass for first particle in a two body mode",
     &GenericWidthGenerator::_MEmass1,
     GeV, 0, 0*GeV, 0*GeV,  1.E12*GeV, false, false, true);

  static ParVector<GenericWidthGenerator,Energy> interfaceMEmass2
    ("MEmass2",
     "The mass for second particle in a two body mode",
     &GenericWidthGenerator::_MEmass2,
     GeV, 0, 0*GeV, 0*GeV,  1.E12*GeV, false, false, true);

  static ParVector<GenericWidthGenerator,Energy> interfaceInterpolationMasses
    ("InterpolationMasses",
     "The masses for interpolation table",
     &GenericWidthGenerator::_intermasses,
     GeV, 0, 0*GeV, 0*GeV,  1E12*GeV, false, false, true);

  static ParVector<GenericWidthGenerator,Energy> interfaceInterpolationWidths
    ("InterpolationWidths",
     "The widths for interpolation table",
     &GenericWidthGenerator::_interwidths,
     GeV, 0, 0*GeV, 0*GeV,  1E12*GeV, false, false, true);

  static ParVector<GenericWidthGenerator,int> interfacenoofenteries
    ("NumberofEntries",
     "The number of entries in the table after this mode",
     &GenericWidthGenerator::_noofentries,
     0, 0, 0, 0, 100000000, false, false, true);

  static Switch<GenericWidthGenerator,bool> interfaceBRNormalize
    ("BRNormalize",
     "Normalize the partial widths so that they have the value BR*Total Width"
     " for an on-shell particle",
     &GenericWidthGenerator::_BRnorm, false, false, false);
  static SwitchOption interfaceBRNormalizeNormalize
    (interfaceBRNormalize,
     "Yes",
     "Perform the normalization",
     true);
  static SwitchOption interfaceBRNormalizeNoNormalisation
    (interfaceBRNormalize,
     "No",
     "Do not perform the normalization",
     false);

  static Parameter<GenericWidthGenerator,double> interfaceBRMinimum
    ("BRMinimum",
     "Minimum branching ratio for inclusion in the running width calculation.",
     &GenericWidthGenerator::_BRminimum, 0.01, 0.0, 1.0,
     false, false, true);

  static Parameter<GenericWidthGenerator,double> interfacePrefactor
    ("Prefactor",
     "The prefactor to get the correct on-shell width",
     &GenericWidthGenerator::_prefactor, 1.0, 0., 1000.,
     false, false, false);

  static Parameter<GenericWidthGenerator,int> interfacePoints
    ("Points",
     "Number of points to use for interpolation tables when needed",
     &GenericWidthGenerator::_npoints, 50, 5, 1000,
     false, false, true);

  static ParVector<GenericWidthGenerator,string> interfaceDecayModes
    ("DecayModes",
     "The tags for the decay modes used in the width generator",
     &GenericWidthGenerator::_decaytags, -1, "", "", "",
     false, false, Interface::nolimits);

  static Parameter<GenericWidthGenerator,unsigned int> interfaceInterpolationOrder
    ("InterpolationOrder",
     "The interpolation order for the tables",
     &GenericWidthGenerator::_intorder, 1, 1, 5,
     false, false, Interface::limited);

}

Energy GenericWidthGenerator::width(const ParticleData &, Energy m) const {
  Energy gamma= Energy();
  for(unsigned int ix =0;ix<_MEcoupling.size();++ix) {
    if(_modeon[ix]) gamma +=partialWidth(ix,m);
  }
  return gamma*_prefactor;
}

void GenericWidthGenerator::doinit() throw(InitException) {
  WidthGenerator::doinit();
  // make sure the particle data object was initialized
  _theParticle->init();
  tDecayIntegratorPtr decayer;
  // mass of the decaying particle
  _mass = _theParticle->mass();
  if(_initialize) {
    // the initial prefactor
    _prefactor=1.;
    // resize all the storage vectors
    _MEtype.clear();
    _MEcode.clear();
    _MEmass1.clear();
    _MEmass2.clear();
    _MEcoupling.clear(); 
    _modeon.clear();
    _minmass.clear();
    _intermasses.clear();
    _interwidths.clear();
    _noofentries.clear();
    _decaytags.clear();
    // integrators that we may need
    WidthCalculatorBasePtr widthptr;
    // get the list of decay modes as a decay selector
    DecayMap modes=_theParticle->decaySelector();
    DecayMap::const_iterator start=modes.begin();
    DecayMap::const_iterator end=modes.end();
    tPDPtr part1,part2;
    tGenericMassGeneratorPtr massgen1,massgen2;
    // loop over the decay modes to get the partial widths
    for(;start!=end;++start) {
      // the decay mode
      tcDMPtr mode=(*start).second;      
      _decaymodes.push_back(const_ptr_cast<DMPtr>(mode));
      _decaytags.push_back(_decaymodes.back()->tag());
      ParticleMSet::const_iterator pit(mode->products().begin());
      // minimum mass for the decaymode
      Energy minmass = Energy();
      for(;pit!=mode->products().end();++pit) {
	(**pit).init();
	minmass+=(**pit).massMin();
      }
      _minmass.push_back(minmass);
      pit=mode->products().begin();
      // its decayer
      decayer=dynamic_ptr_cast<tDecayIntegratorPtr>(mode->decayer());
      if(decayer) decayer->init();
      // if there's no decayer then set the partial width to the br times the
      // on-shell value
      if(!decayer) {
	_MEtype.push_back(0);
	_MEcode.push_back(0);
	_MEcoupling.push_back(mode->brat());
	_MEmass1.push_back(0.*GeV);
	_MEmass2.push_back(0.*GeV);
	_noofentries.push_back(_intermasses.size());
	_modeon.push_back(mode->brat()>_BRminimum);
	setupMode(mode,decayer,_MEtype.size()-1);
      }
      else if(mode->products().size()==2) {
	// the outgoing particles
	ParticleMSet::const_iterator pit = mode->products().begin();
	part1=*pit;++pit;
	part2=*pit;
	// mass generators
	if( part1->stable() || part1->massGenerator())
	  massgen1=dynamic_ptr_cast<tGenericMassGeneratorPtr>(part1->massGenerator());
	else
	  massgen1=tGenericMassGeneratorPtr();
	if(part2->stable() || part2->massGenerator())
	  massgen2=dynamic_ptr_cast<tGenericMassGeneratorPtr>(part2->massGenerator());
	else
	  massgen2=tGenericMassGeneratorPtr();
	if(massgen1) massgen1->init();
	if(massgen2) massgen2->init();
	double coupling(0.);
	int mecode(-1);
	bool order(decayer->twoBodyMEcode(*mode,mecode,coupling));
	_MEcode.push_back(mecode);
	_MEcoupling.push_back(coupling);
	_modeon.push_back(mode->brat()>_BRminimum);
	if(order) {
	  _MEmass1.push_back(part1->mass());
	  _MEmass2.push_back(part2->mass());
	}
	else {
	  _MEmass1.push_back(part2->mass());
	  _MEmass2.push_back(part1->mass());
	}
	// perform setup in the inheriting class
	setupMode(mode,decayer,_MEcode.size()-1);
	// both particles on shell
	if(!massgen1&&!massgen2) {
	  _MEtype.push_back(1);
	  _noofentries.push_back(_intermasses.size());
	  if(_BRnorm) {
	    if(_mass>_MEmass1[_MEtype.size()-1]+_MEmass2[_MEtype.size()-1]) {
	      Energy gamma(partial2BodyWidth(_MEtype.size()-1,_mass));
	      double ratio(mode->brat()*mode->parent()->width()/gamma);
	      ratio=sqrt(ratio);
	      _MEcoupling.back() *=ratio;
	    }
	  }
	}
	else {
	  // one off-shell particle
	  if(!massgen1||!massgen2) {
	    // create the width calculator
	    tGenericWidthGeneratorPtr 
	      ttthis(const_ptr_cast<tGenericWidthGeneratorPtr>(this));
	    WidthCalculatorBasePtr twobody
	      (new_ptr(TwoBodyAllOnCalculator(ttthis,_MEcode.size()-1,
					      _MEmass1[_MEcode.size()-1],
					      _MEmass2[_MEcode.size()-1])));
	    int ioff = ((part1->massGenerator()&&!order)||
			(part2->massGenerator()&&order)) ? 2 : 1;
	    if(massgen1)
	      widthptr=new_ptr(OneOffShellCalculator(ioff,twobody,massgen1,0.*GeV));
	    else
	      widthptr=new_ptr(OneOffShellCalculator(ioff,twobody,massgen2,0.*GeV));
	  }
	  else {
	    int ioff   = order ? 1 : 2;
	    int iother = order ? 2 : 1;
	    // create the width calculator
	    tGenericWidthGeneratorPtr 
	      ttthis(const_ptr_cast<tGenericWidthGeneratorPtr>(this));
	    // this is the both on-shell case
	    WidthCalculatorBasePtr twobody
	      (new_ptr(TwoBodyAllOnCalculator(ttthis,_MEcode.size()-1,
					      _MEmass1[_MEcode.size()-1],
					      _MEmass2[_MEcode.size()-1])));
	    // this is the first off-shell
	    WidthCalculatorBasePtr widthptr2=
	      new_ptr(OneOffShellCalculator(ioff,twobody,massgen1,0*GeV));
	    widthptr=new_ptr(TwoOffShellCalculator(iother,widthptr2,massgen2,
						   0*GeV,massgen1->lowerLimit()));
	  }
	  // set up the interpolation table
	  Energy test(part1->massMin()+part2->massMin());
	  Energy min(max(_theParticle->massMin(),test)),upp(_theParticle->massMax());
	  Energy step((upp-min)/(_npoints-1));
	  Energy moff(min);
	  Energy2 moff2;
	  // additional points to improve the interpolation
	  if(min==test) {
	    _intermasses.push_back(moff-2.*step);_interwidths.push_back(0*GeV);
	    _intermasses.push_back(moff-   step);_interwidths.push_back(0*GeV);
	    _intermasses.push_back(moff        );_interwidths.push_back(0*GeV);
	    double fact(exp(0.1*log(1.+step/moff)));
	    for(unsigned int ix=0;ix<10;++ix) {
	      moff*=fact;
	      moff2=sqr(moff);
	      _intermasses.push_back(moff);
	      _interwidths.push_back(widthptr->partialWidth(moff2));
	    }
	    moff+=step;
	  }
	  else if(test>min-2.*step) {
	    _intermasses.push_back(moff-2.*step);_interwidths.push_back(0*GeV);
	    _intermasses.push_back(test        );_interwidths.push_back(0*GeV);
	  }
	  else {
	    _intermasses.push_back(moff-2.*step);
	    _interwidths.push_back(widthptr->partialWidth((moff-2.*step)*
							  (moff-2.*step)));
	    _intermasses.push_back(moff-   step);
	    _interwidths.push_back(widthptr->partialWidth((moff-   step)*
							  (moff-   step)));
	  }
	  for(; moff<upp+2.5*step;moff+=step) {
	    moff2=moff*moff;
	    _intermasses.push_back(moff);
	    _interwidths.push_back(widthptr->partialWidth(moff2));
	  }
	  coupling=1.;
	  if(_BRnorm) {
	    double ratio(1.);
	    if((massgen1&&massgen2&&
		_mass>massgen1->lowerLimit()+massgen2->lowerLimit())||
	       (massgen1&&!massgen2&&
		_mass>massgen1->lowerLimit()+part2->mass())||
	       (massgen2&&!massgen1&&
		_mass>massgen2->lowerLimit()+part1->mass())||
	       (!massgen1&&!massgen2&&
		_mass>part1->mass()+part2->mass())) {
	      Energy gamma(widthptr->partialWidth(_mass*_mass));
	      ratio=mode->brat()*mode->parent()->width()/gamma;
	    }
	    _MEcoupling.back()=ratio;
	  }
	  else _MEcoupling.back()=1.;
	  _MEtype.push_back(2);
	  _MEcode.back()=0;
	  unsigned int ix=0;
	  if(_MEtype.size()>1){ix=_noofentries[_MEtype.size()-2];}
	  _noofentries.push_back(_intermasses.size());
	  _interpolators.resize(_MEtype.size());
	  // get the vectors we will need
	  vector<Energy>::iterator istart= _intermasses.begin();
	  if(_MEtype.size()>1){istart+=_noofentries[_MEtype.size()-2];}
	  vector<Energy>::iterator iend=_intermasses.end();
	  vector<Energy> masses(istart,iend);

	  istart= _interwidths.begin();
	  if(_MEtype.size()>1){istart+=_noofentries[_MEtype.size()-2];}
	  iend=_interwidths.end();
	  vector<Energy> widths(istart,iend);
	  _interpolators.back() = make_InterpolatorPtr(widths,masses,_intorder);
	}
      }
      // higher multiplicities
      else {
	setupMode(mode,decayer,_MEcode.size());
	widthptr=decayer->threeBodyMEIntegrator(*mode);
	if(!widthptr) {
	  _MEtype.push_back(0);
	  _MEcode.push_back(0);
	  _MEcoupling.push_back(mode->brat());
	  _MEmass1.push_back(0.*GeV);
	  _MEmass2.push_back(0.*GeV);
	  _noofentries.push_back(_intermasses.size());
	  _modeon.push_back(mode->brat()>_BRminimum);
	}
	else {
	  Energy step((_theParticle->widthUpCut()+_theParticle->widthLoCut())/
		      (_npoints-1));
	  Energy moff(_theParticle->massMin()),upp(_theParticle->massMax());
	  for( ; moff<upp+0.5*step;moff+=step) {
	    Energy2 moff2=sqr(moff);
	    Energy wtemp=widthptr->partialWidth(moff2);
	    _intermasses.push_back(moff);
	    _interwidths.push_back(wtemp);
	  }
	  double coupling(1.);
	  if(_BRnorm) {
	    Energy gamma = Energy();
	    gamma=widthptr->partialWidth(_mass*_mass);
	    double ratio(mode->brat()*mode->parent()->width()/gamma);
	    coupling *=ratio;
	  }
	  _MEtype.push_back(2);
	  _MEcode.push_back(0);
	  _MEcoupling.push_back(coupling);
	  _MEmass1.push_back(0*GeV);
	  _MEmass2.push_back(0*GeV);
	  _modeon.push_back(mode->brat()>_BRminimum);
	  unsigned int ix=0;
	  if(_MEtype.size()>1){ix=_noofentries[_MEtype.size()-2];}
	  _noofentries.push_back(_intermasses.size());
	  _interpolators.resize(_MEtype.size());
	  // get the vectors we will need
	  vector<Energy>::iterator istart( _intermasses.begin()),
	    iend(_intermasses.end());
	  if(_MEtype.size()>1){istart+=_noofentries[_MEtype.size()-2];}
	  vector<Energy> masses(istart,iend);
	  
	  istart= _interwidths.begin();
	  if(_MEtype.size()>1){istart+=_noofentries[_MEtype.size()-2];}
	  iend=_interwidths.end();
	  vector<Energy> widths(istart,iend);
	  _interpolators.back() = make_InterpolatorPtr(widths,masses,_intorder);
	}
      }
    }
    // now check the overall normalisation of the running width
    Energy gamma = width(*_theParticle,_mass);
    if(gamma>Energy()) _prefactor = _theParticle->width()/gamma;
    // output the info so it can be read back in
  }
  else {
    // get the decay modes from the tags
    if(_decaytags.size()!=0) {
      _decaymodes.clear();
      for(unsigned int ix=0;ix<_decaytags.size();++ix) {
	_decaymodes.push_back(CurrentGenerator::current().findDecayMode(_decaytags[ix]));
	if(!_decaymodes.back()) 
	  generator()->log() << "Error in GenericWidthGenerator::doinit() "
			     << "Failed to find DecayMode  for tag" 
			     << _decaytags[ix] << "\n";
      }
    }
    // otherwise just use the modes from the selector
    else {
      DecayMap modes(_theParticle->decaySelector());
      DecayMap::const_iterator start(modes.begin()),end(modes.end());
      tcDMPtr mode;
      for(;start!=end;++start) {
	mode=(*start).second;      
	_decaymodes.push_back(const_ptr_cast<DMPtr>(mode));
      }
    }
    // set up the interpolators
    setInterpolators();
  }
  // setup the partial widths in the decayers for normalization
  tDecayIntegratorPtr temp;
  for(unsigned int ix=0;ix<_decaymodes.size();++ix) {
    if(!_decaymodes[ix]) continue;
    _decaymodes[ix]->init();
    decayer=dynamic_ptr_cast<tDecayIntegratorPtr>(_decaymodes[ix]->decayer());
    if(!decayer) continue;
    decayer->init();
    decayer->setPartialWidth(*_decaymodes[ix],ix);
  }
  // code to output plots
//   string fname = CurrentGenerator::current().filename() + 
//     string("-") + name() + string(".top");
//   ofstream output(fname.c_str());
//   Energy step = (_theParticle->massMax()-_theParticle->massMin())/100.;
//   output << "SET FONT DUPLEX\n";
//   output << "TITLE TOP \"Width for " << _theParticle->name() << "\"\n";
//   output << "TITLE BOTTOM \"m/GeV\"\n";
//   output << "TITLE LEFT \"G/GeV\"\n";
//   output << "CASE       \"F    \"\n";
//   output << "SET LIMITS X " 
// 	 << (_theParticle->massMin()-10.*step)/GeV << " " 
// 	 << _theParticle->massMax()/GeV << "\n";
//   Energy upper(0.*GeV);
//   for(Energy etest=_theParticle->massMin();etest<_theParticle->massMax();etest+=step) {
//     Energy gamma=width(*_theParticle,etest);
//     upper = max(gamma,upper);
//     output << etest/GeV << "\t" << gamma/GeV << "\n";
//   }
//   output << "SET LIMITS Y 0. " << upper/GeV << "\n";
//   output << "JOIN\n";
//   output << (_theParticle->massMin()-9.*step)/GeV << "\t" 
// 	 <<  upper*(_MEcode.size()+1)/(_MEcode.size()+2)/GeV << "\n";
//   output << (_theParticle->massMin()-7.*step)/GeV << "\t" 
// 	 <<  upper*(_MEcode.size()+1)/(_MEcode.size()+2)/GeV << "\n";
//   output << "JOIN\n";
//   output << "TITLE DATA " 
// 	 << (_theParticle->massMin()-6.*step)/GeV << "\t" 
// 	 <<  upper*(_MEcode.size()+1)/(_MEcode.size()+2)/GeV 
// 	 << " \"total\"\n";
//   for(unsigned int ix=0;ix<_MEcode.size();++ix) {
//     for(Energy etest=_theParticle->massMin();etest<_theParticle->massMax();etest+=step) {
//       output << etest/GeV << "\t" << partialWidth(ix,etest)*_prefactor/GeV << "\n";
//     }
//     switch(ix) {
//     case 0:  output << "join red\n"    ; break;
//     case 1:  output << "join blue\n"   ; break;
//     case 2:  output << "join green\n"  ; break;
//     case 3:  output << "join yellow\n" ; break;
//     case 4:  output << "join magenta\n"; break;
//     case 5:  output << "join cyan\n"   ; break;
//     case 6:  output << "join dashes\n" ; break;
//     case 7:  output << "join dotted\n" ; break;
//     case 8:  output << "join dotdash\n"; break;
//     default: output << "join daashes space\n";  break;
//     }
//     output << (_theParticle->massMin()-9.*step)/GeV << "\t" 
// 	   <<  upper*(_MEcode.size()-ix)/(_MEcode.size()+2)/GeV << "\n";
//     output << (_theParticle->massMin()-7.*step)/GeV << "\t" 
// 	   <<  upper*(_MEcode.size()-ix)/(_MEcode.size()+2)/GeV << "\n"; 
//     switch(ix) {
//     case 0:  output << "join red\n"    ; break;
//     case 1:  output << "join blue\n"   ; break;
//     case 2:  output << "join green\n"  ; break;
//     case 3:  output << "join yellow\n" ; break;
//     case 4:  output << "join magenta\n"; break;
//     case 5:  output << "join cyan\n"   ; break;
//     case 6:  output << "join dashes\n" ; break;
//     case 7:  output << "join dotted\n" ; break;
//     case 8:  output << "join dotdash\n"; break;
//     default: output << "join daashes space\n";  break;
//     }
//     output << "TITLE DATA " 
// 	   << (_theParticle->massMin()-6.*step)/GeV << "\t" 
// 	   <<  upper*(_MEcode.size()-ix)/(_MEcode.size()+2)/GeV 
// 	   << " \"" << _decaytags[ix] << "\"\n";
//   }
}
 
void GenericWidthGenerator::setInterpolators() {
  // create the interpolators
  _interpolators.resize(_MEtype.size());
  vector<Energy>::iterator estart(_intermasses.begin()),eend;
  vector<Energy>::iterator wstart(_interwidths.begin()),wend;
  vector<Energy> masses,widths;
  for(unsigned int ix=0;ix<_MEtype.size();++ix) {
    eend=_intermasses.begin()+_noofentries[ix];
    wend=_interwidths.begin()+_noofentries[ix];
    if(_MEtype[ix]==2) {
      masses.assign(estart,eend);
      widths.assign(wstart,wend);
      _interpolators[ix]= make_InterpolatorPtr(widths,masses,_intorder);
    }
    estart=eend;
    wstart=wend;
  }
}

void GenericWidthGenerator::dataBaseOutput(ofstream & output, bool header) {
  if(header) output << "update Width_Generators set parameters=\"";
  // prefactor and general switiches
  output << "set " << name() << ":Prefactor "   << _prefactor << "\n";
  output << "set " << name() << ":BRNormalize " << _BRnorm    << "\n";
  output << "set " << name() << ":BRMinimum "   << _BRminimum << "\n";
  output << "set " << name() << ":Points "      << _npoints   << "\n";
  output << "set " << name() << ":InterpolationOrder " << _intorder << "\n";
  // the type of the matrix element
  for(unsigned int ix=0;ix<_MEtype.size();++ix) {
    output << "insert " << name() << ":MEtype " << ix << " " 
	   << _MEtype[ix] << "\n";
  }
  // the code for thew two body matrix elements
  for(unsigned int ix=0;ix<_MEcode.size();++ix) {
    output << "insert " << name() << ":MEcode " 
	   << ix << " " << _MEcode[ix] << "\n";
  }
  // the coupling for trhe two body matrix elements
  for(unsigned int ix=0;ix<_MEcoupling.size();++ix) {
    output << "insert " << name() << ":MEcoupling " 
	   << ix << " " << _MEcoupling[ix] << "\n";
  }
  // use this mode for the running width
  for(unsigned int ix=0;ix<_modeon.size();++ix) {
    output << "insert " << name() << ":ModeOn " 
	   << ix << " " << _modeon[ix] << "\n";
  }
  // first outgoing mass
  for(unsigned int ix=0;ix<_minmass.size();++ix) {
    output << "insert " << name() << ":MinimumMasses " 
	   << ix << " " << _minmass[ix]/GeV << "\n";
  }
  // first outgoing mass
  for(unsigned int ix=0;ix<_MEmass1.size();++ix) {
    output << "insert " << name() << ":MEmass1 " 
	   << ix << " " << _MEmass1[ix]/GeV << "\n";
  }
  // second outgoing mass
  for(unsigned int ix=0;ix<_MEmass2.size();++ix) {
    output << "insert " << name() << ":MEmass2 " 
	   << ix << " " << _MEmass2[ix]/GeV << "\n";
  }
  for(unsigned int ix=0;ix<_decaymodes.size();++ix) {
    output << "insert " << name() << ":DecayModes "
	   << ix << " " << _decaytags[ix] << " \n";
  }
  // data for the interpolation tables
  std::streamsize curpre=output.precision();
  output.precision(curpre+2);
  for(unsigned int ix=0;ix<_intermasses.size();++ix) {
    output << "insert " << name() 
	   << ":InterpolationMasses " 
	   << ix << " " << _intermasses[ix]/GeV << "\n";
  }
  for(unsigned int ix=0;ix<_interwidths.size();++ix) {
    output << "insert " << name() 
	   << ":InterpolationWidths " 
	   << ix << " " << _interwidths[ix]/GeV << "\n";
  }
  output.precision(curpre);
  for(unsigned int ix=0;ix<_noofentries.size();++ix) {
    output << "insert " << name() 
	   << ":NumberofEntries " 
	   << ix << " " << _noofentries[ix] << "\n";
  }  
  if(header) output << "\n\" where BINARY ThePEGName=\"" << name() 
		    << "\";" << endl;
}

DecayMap GenericWidthGenerator::rate(const Particle & p) {
  // return default if not using running widths
  if(!_theParticle->variableRatio()) return p.data().decaySelector();
  // use the running widths to generate the branching ratios
  Energy scale(p.mass());
  DecayMap dm;
  Energy width = _theParticle->width();
  for(unsigned int ix=0;ix<_decaymodes.size();++ix) {
    dm.insert(partialWidth(ix,scale)/width,
	      p.id()==_theParticle->id() ? 
	      _decaymodes[ix] : _decaymodes[ix]->CC());
  }
  return dm;
}

void GenericWidthGenerator::setupMode(tcDMPtr, tDecayIntegratorPtr,
				      unsigned int)
{}

Energy GenericWidthGenerator::partialWidth(int imode,Energy q) const {
  if(q<_minmass[imode]) return Energy();
  Energy gamma;
  if(_MEtype[imode]==0) {
    gamma=_MEcoupling[imode]*_theParticle->width();
  }
  else if(_MEtype[imode]==1) {
    gamma=partial2BodyWidth(imode,q);
  }
  else if(_MEtype[imode]==2) {
    gamma=_MEcoupling[imode]*(*_interpolators[imode])(q);
  }
  else {
    throw Exception() << "Unknown type of mode " << _MEtype[imode] 
		      << "in GenericWidthGenerator::partialWidth()"
		      << Exception::runerror;
  }
  return max(gamma,Energy());
}

void GenericWidthGenerator::dofinish() {
  if(_initialize) {
    string fname = CurrentGenerator::current().filename() + 
      string("-") + name() + string(".output");
    ofstream output(fname.c_str());
    dataBaseOutput(output,true);
  }
  WidthGenerator::dofinish();
}

void GenericWidthGenerator::doinitrun() {
  WidthGenerator::doinitrun();
  // set up the interpolators
  setInterpolators();
}

void GenericWidthGenerator::rebind(const TranslationMap & trans)
  throw(RebindException) {
  _theParticle = trans.translate(_theParticle);
  WidthGenerator::rebind(trans);
}

IVector GenericWidthGenerator::getReferences() {
  IVector ret = WidthGenerator::getReferences();
  ret.push_back(_theParticle);
  return ret;
}

Length GenericWidthGenerator::lifeTime(const ParticleData &, Energy m, Energy w) const {
  if(m<_theParticle->massMin()) w = width(*_theParticle,_theParticle->massMin());
  else if(m>_theParticle->massMax()) w = width(*_theParticle,_theParticle->massMax());
  return UseRandom::rndExp(hbarc/w);
}

Energy GenericWidthGenerator::partial2BodyWidth(int imode, Energy q,Energy m1,
						       Energy m2) const {
  using Constants::pi;
  if(q<m1+m2) return Energy();
  // calcluate the decay momentum
  Energy2 q2(q*q),m02(_mass*_mass),m12(m1*m1),m22(m2*m2),
    pcm2(0.25*(q2*(q2-2.*m12-2.*m22)+(m12-m22)*(m12-m22))/q2);
  if(_MEcode[imode]==-1) return q/_mass*_theParticle->width();
  Energy  pcm(sqrt(pcm2));
  double gam(0.);
  switch(_MEcode[imode]) {
  // V -> P P
  case  0: gam = pcm2/6./q2;
    break;
  // V -> P V
  case  1: gam = pcm2/12./m02;
    break;
  // V -> f fbar
  case  2: gam = 1./12.*(q2*(2.*q2-m12-m22+6.*m1*m2)
			 -(m12-m22)*(m12-m22))/q2/q2;
    break;
  // P -> VV
  case  3: gam = 0.25*pcm2/m02;
    break;
  // A -> VP 
  case  4: gam = (2.*pcm2+3.*m12)/24./m02;
    break;
  // V -> VV
  case  5: gam = pcm2/3./q2*(1.+m12/q2+m22/q2);
    break;
  // S -> SS
  case  6: gam = 0.125/q2*m02;
    break;
  // T -> PP
  case  7: gam = pcm2*pcm2/60./q2/m02;
    break;
  // T -> VP
  case  8: gam = pcm2*pcm2/40./m02/m02;
    break;
  // T -> VV
  case  9: gam = 1./30./q2/q2/m02*
      (3.*q2*(8.*pcm2*pcm2+5.*(m12*m22+pcm2*(m12+m22)))
       -5.*(m12-m22)*(m12-m22)*pcm2);
    break;
  // P -> PV
  case 10: gam = 0.5*pcm2/m22;
    break;
  // P -> PT
  case 11: gam = sqr(pcm2)/12.*q2/m12/m12/m02;
    break;
  // S -> VV
  case 12: gam = 0.125*(2.*pcm2+3.*m12*m22/q2)/m02;
    break;
  // unknown
  default:
    throw Exception() << "Unknown type of mode " << _MEcode[imode] 
		      << " in GenericWidthGenerator::partial2BodyWidth() " 
		      << Exception::abortnow;
  }
  return gam*pcm*sqr(_MEcoupling[imode])/pi;
}
